/*
    Miguel Ramirez, Nir Lipovetzky, Hector Geffner
    C^3: A planner for the sequential, satisficing track of the IPC-6
    Copyright (C) 2008  

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <iostream>
#include <cstdlib>
#include <fstream>
#include <signal.h>
#include "utils.hxx"
#include "PDDL.hxx"
#include "nff_options.hxx"
#include "nff_h2.hxx"
#include "nff_lazy_jp_propagator.hxx"
#include "nff_state.hxx"
#include "nff_c3_bfs.hxx"
#include "nff_hc.hxx"
#include "nff_h_add.hxx"


template <typename Solver_Type>
void extract_common_dfs_stats( Solver_Type& s, float t0, float tf, std::ofstream& stats )
{
	stats << "Max_Label_Size=" << s.max_label_size << std::endl;
	stats << "Avg_Label_Size=" << s.avg_label_size << std::endl;
	stats << "Avg_Branching_Factor=" << s.avg_branching_factor << std::endl; 
	stats << "HC_Time="; report_interval( t0, tf, stats );
	stats << "HC_Expanded=" << s.expanded << std::endl;
}

template <typename BFS_Solver_Type>
void extract_bfs_stats( BFS_Solver_Type& s, float t0, float tf, std::ofstream& stats )
{
	stats << "BFS_Time="; report_interval( t0, tf, stats );
	stats << "BFS_Expanded=" << s.expanded << std::endl;
}

template <typename Propagator_Type>
bool compute_plan_dfs( PDDL::Task& task, std::ofstream& stats )
{
	NFF::Hill_Climbing< Propagator_Type > planner;

	float t0, tf;
	t0 = time_used();
	bool solved = planner.solve();
	tf = time_used();

	extract_common_dfs_stats( planner, t0, tf, stats );

	if ( solved )
	{
		std::cout << "Problem solved. Plan has " << planner.path_found.size() << " actions" << std::endl;
		stats << "HC_Plan=Yes" << std::endl;
		stats << "BFS_Time=n/a" << std::endl;
		stats << "BFS_Expanded=n/a" << std::endl;
		stats << "BFS_Plan=No" << std::endl;
		stats << "Nr_Actions=" << planner.path_found.size() << std::endl;
		// Output plan
		std::ofstream plan_out( "plan.solution" );
		unsigned action_count = 1;
		float plan_cost = 0.0f;
		for ( unsigned k = 0; k < planner.path_found.size(); k++ )
		{
			PDDL::Operator* op = task.useful_ops()[planner.path_found[k]];
			plan_out << (float)action_count++ << ": ";
			task.print_operator( op, plan_out );
			plan_cost = std::add( plan_cost, task.op_cost(planner.path_found[k]) );
			plan_out << std::endl;
		}
		std::cout << "Metric: Total cost of plan: " << plan_cost << std::endl;
		stats << "Cost=" << plan_cost << std::endl;
		plan_out.close();
	}
	else
	{
		std::cout << "HC did not solve the problem :(" << std::endl;
		stats << "HC_Plan=No" << std::endl;
	}
	return solved;	
}

bool compute_plan_bfs( PDDL::Task& task, std::ofstream& stats )
{
	NFF::C3_Best_First_Search planner;

	float t0, tf;
	t0 = time_used();
	bool solved = planner.solve();
	tf = time_used();

	extract_bfs_stats( planner, t0, tf, stats );
	if ( solved )
	{
		std::cout << "Problem solved. Plan has " << planner.path_found.size() << " actions" << std::endl;
		stats << "BFS_Plan=Yes" << std::endl;
		stats << "Nr_Actions=" << planner.path_found.size() << std::endl;
		// Output plan
		std::ofstream plan_out( "plan.solution" );
		unsigned action_count = 1;
		float plan_cost = 0.0f;
		for ( unsigned k = 0; k < planner.path_found.size(); k++ )
		{
			PDDL::Operator* op = task.useful_ops()[planner.path_found[k]];
			plan_out << (float)action_count++ << ": ";
			task.print_operator( op, plan_out );
			plan_cost = std::add( plan_cost, task.op_cost(planner.path_found[k]) );
			plan_out << std::endl;
		}
		std::cout << "Metric: Total cost of plan: " << plan_cost << std::endl;
		stats << "Cost=" << plan_cost << std::endl;
		plan_out.close();

	}
	else
	{
		std::cout << "BFS did not solve the problem :(" << std::endl;
		stats << "BFS_Plan=No" << std::endl;	
		stats << "Nr_Actions=--" << std::endl;	
	}
	return solved;
}

int main( int argc, char** argv )
{
	double t0, tf;

	NFF_Options::parse_command_line( argc, argv );
	NFF_Options& prog_opts = NFF_Options::instance();

	PDDL::Task& task = PDDL::Task::instance();
	task.setup();


	std::ofstream stats( "execution.stats" );	

	if ( prog_opts.verbose_mode() )
	{
		std::ofstream fluents_out( "fluents.list" );
		std::ofstream ops_out( "operators.list" );
		std::ofstream init_out( "initial.list" );
		std::ofstream goal_out( "goal.list" );
	
		task.print_fluents( fluents_out );
		task.print_operators( ops_out );
		task.print_initial_state( init_out );
		task.print_goal_state( goal_out );

		fluents_out.close();
		ops_out.close();
	}
	stats << "Fluents=" << task.fluents().size() << std::endl;
	std::cout << "Fluents=" << task.fluents().size() << std::endl;
	stats << "Operators=" << task.useful_ops().size() << std::endl;	
	std::cout << "Operators=" << task.useful_ops().size() << std::endl;

	if ( NFF_Options::instance().only_grounding() )
	{
		std::cout << "Stopping after PDDL task has been processed" << std::endl;
		std::exit(0);
	}

	task.set_h2( new NFF::H2 );
	t0 = time_used();
	task.h2().compute_only_mutexes();
	bool solvable = task.h2().extract_op_reachability_info();
	task.h2().compute_e_deletes();
	tf = time_used();
	
	stats << "Preprocessing=";  report_interval( t0, tf, stats );

	if ( !solvable )
	{
		stats << "Plan=No" << std::endl;
		std::cout << "Planning task not solvable" << std::endl;
		std::exit(0);
	}	

	#ifndef NDEBUG
	std::cout << "Initial state: ";
	NFF::State* s0 = NFF::State::make_initial_state();	
	s0->print( std::cout );
	std::cout << std::endl;
	std::cout << "Goal state: ";
	NFF::State* sG = NFF::State::make_goal_state();
	sG->print( std::cout );
	std::cout << std::endl;
	#endif

	typedef NFF::C3_Lazy_JP_Propagator< NFF::Additive_Heuristic >			H_Add_C3_JP_Propagator;


	bool solved_HC = false;
	solved_HC = compute_plan_dfs<H_Add_C3_JP_Propagator>( task, stats );

	if ( !solved_HC )
		compute_plan_bfs( task, stats );
	
	stats.close();
	
	std::exit(0);
	return 0;
}
